from rich.table import Table
from rich.console import Console
from rich.text import Text
from rich.style import Style

from typing import Iterable, Optional, List, Union, Dict, Sequence, Any

from ..model import Organization, Project, Status, Member, WorkItem

def _show_table(table_data: Iterable[Any], headers: Optional[List[str]], title: Optional[str] = None):
    console = Console()
    table = Table(title=title)
    
    # Add headers
    if headers:
        for header in headers:
            table.add_column(header)
    
    # Add rows
    for row in table_data:
        table.add_row(*[cell if isinstance(cell, Text) else str(cell) for cell in row])
    
    console.print(table)

def show_table_generic(items: Iterable[Iterable[Any]], headers=List[str]):
    '''
    Print generic iterable items with Rich table.
    '''
    _show_table(items, headers=headers)

def show_table_organization(items: Iterable[Organization]):
    '''
    Print organizations with Rich table.
    '''

    table = list(map(lambda x: [x.id, x.name, x.isOrgAdmin], items))
    headers = ['ID', 'Name', 'IsAdmin']
    _show_table(table, headers=headers)

def show_table_project(items: Iterable[Project]):
    '''
    Print projects with Rich table.
    '''

    table = list(map(lambda x: [x.identifier, x.customCode, x.name], items))
    headers = ['ID', 'Code', 'Name']
    _show_table(table, headers=headers)

def show_table_project_status(items: Iterable[Status]):
    '''
    Print projects with Rich table.
    '''

    table = list(map(lambda x: [x.identifier, x.name, x.source], items))
    headers = ['ID', 'Name', 'Source']
    _show_table(table, headers=headers)

def show_table_member(items: Iterable[Member]):
    '''
    Print members with Rich table.
    '''

    table = list(map(lambda x: [x.identifier, x.displayName, x.roleName], items))
    headers = ['ID', 'Name', 'Role']
    _show_table(table, headers=headers)

def show_table_workitem(items: Iterable[WorkItem]):
    '''
    Print work items with Rich table.
    '''
    def add_link(text, url):
        txt = Text()
        txt.append(text, style=Style(color="green", link=url))
        return txt
    
    def get_username(user):
        from ..main import GlobalState
        return GlobalState.current().get_matching_member_name(user)

    table = list(map(lambda x: [add_link(x.serialNumber, x.web_url), x.subject, get_username(x.assignedTo), x.status], items))
    headers = ['NO.', 'Title', 'Assigned To', 'Status']
    _show_table(table, headers=headers)

def show_panel_workitem(item: WorkItem):
    '''
    Print work items with Rich table.
    '''
    from ..main import GlobalState

    def add_link(text, url):
        txt = Text()
        txt.append(text, style=Style(color="green", link=url))
        return txt
    
    def get_username(user):
        from ..main import GlobalState
        return GlobalState.current().get_matching_member_name(user)

    def format_timestamp(value):
        from datetime import datetime, timezone
        return datetime.fromtimestamp(value / 1000, tz=timezone.utc).strftime('%Y-%m-%d %H:%M:%S')

    def print_property(title, content):
        console = Console()
        text = Text()
        text.append(title)
        text.append(': ')

        if isinstance(content, Text):
            text.append(content)
        else:
            text.append(content, style="bold italic")

        console.print(text)

    print_property('ID', add_link(item.serialNumber, item.web_url))
    print_property('Subject', item.subject)
    print_property('Project', item.spaceName)
    print_property('Category', item.categoryIdentifier)
    print_property('Assigned To', get_username(item.assignedTo))
    print_property('Status', item.status)
    print_property('Update Status At', format_timestamp(item.updateStatusAt))
    print_property('Creator', get_username(item.creator))
    print_property('Created At', format_timestamp(item.gmtCreate))
    print_property('Modifier', get_username(item.modifier))
    print_property('Modified At', format_timestamp(item.gmtModified))
    print_property('Finish Time', format_timestamp(item.finishTime))

    project_id = item.spaceIdentifier
    category = item.categoryIdentifier

    for field in item.customFields:
        field_name = GlobalState.current().get_matching_field_name(project_id, category, field.get('fieldIdentifier'))
        field_value = item.get_display_value_of_custom_field(field.get('fieldIdentifier'))
        print_property(field_name, field_value)

def show_content(title: str, content: str):
    console = Console()
    text = Text()
    text.append(title)
    text.append(content, style="bold italic")
    console.print(text)



import click

from .sdk import *
from .utils.output import *
from .utils.cache import *
from .utils.date import *
from .utils.command import *
from .utils.state import *
from .web import *

@click.group()
def cli():
    """CLI tool for managing yunxiao projects"""
    pass

@cli.group(invoke_without_command=True)
@click.pass_context
def organization(ctx):
    """Project management commands"""
    if ctx.invoked_subcommand is None:
        ctx.invoke(organization_list)

@organization.command(name='list')
def organization_list():
    """List all the organizations"""
    organizations = OrganizationListAPI.run()
    [save_cached_organization(org) for org in organizations]
    show_table_organization(organizations)

@cli.group(invoke_without_command=True)
@click.pass_context
def project(ctx):
    """Project management commands"""
    if ctx.invoked_subcommand is None:
        ctx.invoke(project_list_builtin)

@project.command(name='list')
def project_list_builtin():
    """List the cached projects"""
    projects = ProjectListAPI.run(GlobalState.current().organization_id)
    show_table_project(projects)

@project.command(name='all')
@click.option('--reload', '-r', is_flag=True, default=False, help='Force reload the following projects or not.')
@click.option('--reload-project-set', '-s', is_flag=True, default=False, help='Force reload the online project sets or not.')
def project_list_all(reload, reload_project_set):
    """List/Reload the builtin and following projects"""
    projects = GlobalState.current().get_all_projects(reload, reload_project_set)
    show_table_project(projects)

@project.command(name='field')
def project_fields_all():
    """List all the projects' field"""
    for project in GlobalState.current().get_all_projects():
        click.echo(f'Project: {project.name}')
        for category, workitemtypes in project.workitemtypes.items():
            click.echo(f'\tCategory: {category}')
            for workitemtype in workitemtypes:
                click.echo(f'\t\tType: {workitemtype.name}')
                # for field in workitemtype.fields:
                #     click.echo(f'\t\t\tField: {field.name}')
                click.echo(f'\t\tFields: {", ".join(sorted(map(lambda x: x.name, workitemtype.fields)))}')
        else:
            click.echo()

@project.command(name='info')
@click.argument('id', type=click.STRING, required=True)
def workitem_info(id):
    """Get workitem details"""
    project = ProjectDetailAPI.run(GlobalState.current().organization_id, id)
    show_table_project([project])

@project.command(name='status')
@click.option('--project', type=click.STRING, help='Project ID')
@click.option('--category', type=click.Choice([c.value for c in Category]), required=True, help='Category name')
def project_status(project, category):
    """List the project's statuses"""
    if project:
        all_projects = [ProjectDetailAPI.run(GlobalState.current().organization_id, project)]
    else: 
        all_projects = GlobalState.current().get_all_projects()

    for project in all_projects:
        statuses = WorkFlowStatusAPI.run(GlobalState.current().organization_id, project.identifier, category)
        if statuses:
            project.statuses = statuses
            GlobalState.current().save_organization()
            show_content('Project: ', project.name)
            show_table_project_status(statuses)

@project.command(name='member')
@click.option('--project', type=click.STRING, help='Project ID')
def member_list(project):
    """List all members under the project"""
    if project:
        all_projects = [ProjectDetailAPI.run(GlobalState.current().organization_id, project)]
    else:
        all_projects = GlobalState.current().get_all_projects()

    for project in all_projects:
        members = MemberListAPI.run(GlobalState.current().organization_id, project.identifier)
        if members:
            project.members = members
            GlobalState.current().save_organization()
            show_content('Project: ', project.name)
            show_table_member(members)

@cli.group(invoke_without_command=True)
@click.pass_context
def workitem(ctx):
    """Workitem management commands"""
    if ctx.invoked_subcommand is None:
        ctx.invoke(workitem_task_list)

@workitem.command(name='task')
@click.option('--assigned-to', type=click.STRING, help='User ID/Name')
@click.option('--statuses', type=click.STRING, help='Workitem statuses, use comma as separator')
@click.option('--not-in-statuses', type=click.STRING, help='Reversed workitem statuses, use comma as separator')
def workitem_task_list(assigned_to, statuses, not_in_statuses):
    """List task workitems with filters"""

    for project, workitems in list_workitem_by(Category.Task, assigned_to, statuses, not_in_statuses):
        if workitems:
            show_content('Project: ' , project.name)
            show_table_workitem(workitems)

@workitem.command(name='bug')
@click.option('--assigned-to', type=click.STRING, help='User ID/Name')
@click.option('--statuses', type=click.STRING, help='Workitem statuses, use comma as separator')
@click.option('--not-in-statuses', type=click.STRING, help='Reversed workitem statuses, use comma as separator')
def workitem_bug_list(assigned_to, statuses, not_in_statuses):
    """List bug workitems with filters"""
    for project, workitems in list_workitem_by(Category.Bug, assigned_to, statuses, not_in_statuses):
        if workitems:
            show_content('Project: ' , project.name)
            show_table_workitem(workitems)

@workitem.command(name='info')
@click.argument('id', type=click.STRING, required=True)
def workitem_info(id):
    """Get workitems details"""

    org = GlobalState.current().organization_id
    # Get the basic info of work item w/o custom fields
    workitem = WorkItemDetailAPI.run(org, id)
    # Get the basic info of work item with custom fields
    workitem = WorkItemDetailAPI.run(org, workitem.identifier)

    show_panel_workitem(workitem)

@cli.command(name='test', hidden=True)
def test_entry():
    """Test"""
    pass

@cli.group(invoke_without_command=True)
@click.pass_context
def do(ctx):
    '''
    Do something magic.
    '''

    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())

def inject_do_commands():
    '''
    Inject the custom plugin commands definded in configuration file to
    `do` group.
    '''

    def bind(cls):
        def func(**kwargs):
            logging.debug(f'Start running plugin [{cls.name()}]')
            cls().run(**kwargs)
            logging.debug(f'Finish running plugin [{cls.name()}]')

        func.__name__ = cls.name()
        return func

    for cls in get_user_commands():
        command = click.Command(
            cls.name(),
            callback=bind(cls),
            help=cls.help(),
            params=cls.arguments() + cls.options(),
        )
        do.add_command(command)

inject_do_commands()

if __name__ == '__main__':
    cli()



from typing import Dict, List, Union

class Condition:
    def __init__(self, dict) -> None:
        self._dict = dict

    @classmethod
    def assigned_to(cls, value: str) -> 'Condition':
        """Initialize a condition for assigned_to field"""
        return cls({
            "fieldIdentifier": "assignedTo",
            "operator": "CONTAINS",
            "value": [value],
            "toValue": None,
            "className": "user",
            "format": "list"
        })

    @classmethod
    def status_contains(cls, value: Union[str, List[str]]) -> 'Condition':
        """Initialize a condition for status field"""
        return cls({
            "fieldIdentifier": "status",
            "operator": "CONTAINS",
            "value": [value] if isinstance(value, str) else value,
            "toValue": None,
            "className": "status",
            "format": "list"
        })

    @classmethod
    def status_not_contains(cls, value: Union[str, List[str]]) -> 'Condition':
        """Initialize a condition for status field"""
        return cls({
            "fieldIdentifier": "status",
            "operator": "NOT_CONTAINS",
            "value": [value] if isinstance(value, str) else value,
            "toValue": None,
            "className": "status",
            "format": "list"
        })

    @property
    def dict(self) -> Dict:
        return self._dict

class ConditionGroup:
    def __init__(self) -> None:
        self._conditions = []

    def add(self, condition: Condition) -> None:
        self._conditions.append(condition)

    @property
    def dict(self) -> Dict:
        return {
            "conditionGroups": [
                list(map(lambda x: x.dict, self._conditions))
            ]
        }



from .base import *

class Member(ModelBase):
    '''
    {
        'avatar': 'https://tcs-devops.aliyuncs.com/thumbnail/ffffffff/w/100/h/100',
        'displayName': 'GOD',
        'division': {},
        'identifier': '1234567890',
        'nickName': 'GOD',
        'organizationUserInfo': {},
        'realName': 'GOD',
        'roleName': '管理员',
        'tbRoleId': 'project.admin'
    }
    '''

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Member):
            return False
        return self.identifier == other.identifier



from typing import Optional, List, Any
from dataclasses import dataclass
from os.path import exists, dirname, basename
import sys
import importlib
from abc import ABC, abstractmethod
from click import Parameter

class CommandBase(ABC):
    '''All the custom commands\' base class.'''

    @classmethod
    def name(cls) -> str:
        return ''

    @classmethod
    def help(cls) -> str:
        return ''
    
    @classmethod
    def arguments(cls) -> List[Parameter]:
        return []
    
    @classmethod
    def options(cls) -> List[Parameter]:
        return []

    @abstractmethod
    def run(self, **kwargs):
        """Run the command logic."""
        pass

class CommandCenter:
    '''Manage commands for dispatching.'''

    def __init__(self, commands: List[CommandBase.__subclasses__]):
        self.commands = commands

    @property
    def is_valid(self):
        return len(self.commands) > 0

    @staticmethod
    def register(path: str) -> Optional[CommandBase.__subclasses__]:
        dirpath = dirname(path)
        filename = basename(path)

        if not exists(dirpath):
            return None

        if not filename.endswith('.py'):
            return None

        module_name = filename.replace('.py', '')
        is_new_path = False

        if dirpath not in sys.path:
            sys.path.append(dirpath)
            is_new_path = True

        # Dynamically import the specified module using its filename and directory path
        module = importlib.import_module(module_name, dirpath)
        # Retrieve the Command class from the dynamically imported module
        command = getattr(module, 'Command')

        if is_new_path:
            sys.path.remove(dirpath)

        return command

    def run(self, items: List[str], **kwargs):
        for item in items:
            if exists(item):
                cmd_cls = CommandBase.register(item)

                if cmd_cls:
                    print(f'Runing command: {item}')
                    cmd_cls(**kwargs).run()
                else:
                    print('Didn\'t find a class named Command in the module')
            else:
                print(f'Invalid plugin: {item}')



from abc import ABC
from typing import Any, List, Dict, Optional

class ModelBase(ABC):

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def _format_value(self, value: Any, indent: int = 0) -> str:
        if isinstance(value, dict):
            items = []
            for k, v in sorted(value.items()):
                items.append(f"{'  ' * (indent + 1)}{k}={self._format_value(v, indent + 1)}")
            return "{\n" + "\n".join(items) + f"\n{'  ' * indent}}}"
        elif isinstance(value, (list, tuple)):
            items = []
            for v in value:
                items.append(f"{'  ' * (indent + 1)}{self._format_value(v, indent + 1)}")
            return "[\n" + "\n".join(items) + f"\n{'  ' * indent}]"
        elif isinstance(value, ModelBase):
            return value.__class__.__name__ + "(\n" + "\n".join(
                f"{'  ' * (indent + 1)}{k}={self._format_value(v, indent + 1)}"
                for k, v in sorted(value.__dict__.items())
            ) + f"\n{'  ' * indent})"
        return str(value)

    def __repr__(self):
        items = []
        for key, value in sorted(self.__dict__.items()):
            items.append(f"  {key}={self._format_value(value, 1)}")
        return f"{self.__class__.__name__}(\n" + "\n".join(items) + "\n)"



from .base import *

class WorkItemField(ModelBase):
    '''
    工作项字段

    {
      "identifier": "subject",
      "isRequired": true,
      "name": "标题",
      "format": "input",
      "description": "标题不能为空",
      "className": "string",
      "type": "NativeField"
    }
    {
      "identifier": "assignedTo",
      "isRequired": true,
      "gmtModified": 1728538669000,
      "isShowWhenCreate": true,
      "name": "负责人",
      "format": "list",
      "description": "负责人",
      "isSystemRequired": true,
      "className": "user",
      "gmtCreate": 1728538669000,
      "type": "NativeField",
      "resourceType": "Workitem"
    }
    '''



from .base import *

class WorkItem(ModelBase):
    '''
    {
        'assignedTo': '1234567890',
        'categoryIdentifier': 'Task',
        'creator': '1234567890',
        'document': '',
        'gmtCreate': 1741345082000,
        'gmtModified': 1742377068000,
        'identifier': '1234567890',
        'logicalStatus': 'NORMAL',
        'modifier': '1234567890',
        'parentIdentifier': 'EMPTY_VALUE',
        'serialNumber': 'ABC-123',
        'spaceIdentifier': 'ffffffff',
        'spaceName': 'NAME',
        'spaceType': 'Project',
        'status': '测试中',
        'statusIdentifier': '100012',
        'statusStageIdentifier': '12',
        'subject': 'Setup environment',
        'updateStatusAt': 1742377068000,
        'workitemTypeIdentifier': 'ffffffff',
        'customFields': [
          {
            'fieldClassName': 'option',
            'fieldFormat': 'list',
            'fieldIdentifier': 'priority',
            'value': 'xxxxxxxx',
            'valueList': [{'displayValue': '中',
                          'identifier': 'xxxxxxxx',
                          'level': 4,
                          'value': '中',
                          'valueEn': 'Medium'}],
            'workitemIdentifier': 'xxxxxxxx'}]
          }
        ]
      }
    '''

    @property
    def web_url(self):
        return f'https://devops.aliyun.com/projex/project/{self.spaceIdentifier}/{self.categoryIdentifier.lower()}/{self.identifier}'

    def get_display_value_of_custom_field(self, field_id: str) -> Optional[str]:
        """
        Retrieves the value of a custom field by its identifier.
        """
        for field in self.customFields:
            if field['fieldIdentifier'] == field_id:
                return field.get('valueList', [])[0].get('displayValue')



from os import makedirs, chmod
from os.path import join, exists, expanduser
from typing import Tuple, List, Dict, Optional, Any, Union
import yaml
from textwrap import dedent

from ..model import Project
from .command import CommandCenter, CommandBase

_config = None

def get_config_root() -> str:
    _root = expanduser('~/.config/yunxiao-cli')
    exists(_root) or makedirs(_root, exist_ok=True)  # type: ignore[func-returns-value]
    return _root

def _get_config_file() -> str:
    return join(get_config_root(), 'config.yaml')

def _load_config() -> dict:
    global _config

    if _config is not None:
        return _config

    config_file = _get_config_file()

    if not exists(config_file):
        content = dedent('''
        # User Credential
        # https://help.aliyun.com/zh/ram/user-guide/create-an-accesskey-pair
        CREDENTIAL:
            access_key_id: 
            access_key_secret:

            # Open page https://myaccount.console.aliyun.com/overview and copy the value of "账号ID"
            aliyun_account_id:

        TEAM:
            endpoint: devops.cn-hangzhou.aliyuncs.com
            organization: 

            # Following projects
            projects:
                - id: 
                  name: 
      
        COMMAND:
            - # /path/to/code.py
        ''')

        with open(config_file, 'w') as file:
            file.write(content)
        
        chmod(config_file, 0o600)  # u=rw,g=,o=

    with open(config_file) as file:
        _config = yaml.load(file, Loader=yaml.FullLoader)

    return _config

def get_credential() -> Tuple[str, str]:
    config = _load_config()
    access_key_id = config.get('CREDENTIAL', {}).get('access_key_id')
    access_key_secret = config.get('CREDENTIAL', {}).get('access_key_secret')

    return access_key_id, access_key_secret

def get_aliyun_account_id() -> str:
    return _load_config().get('CREDENTIAL', {}).get('aliyun_account_id')

def get_prefer_organization() -> str:
    return _load_config().get('TEAM', {}).get('organization')

def get_endpoint() -> str:
    return _load_config().get('TEAM', {}).get('endpoint')

def get_following_projects() -> List[Project]:
    data = [{
        'customCode': '',
        'identifier': x.get('id'),
        'name': x.get('name'),
    } for x in _load_config().get('TEAM', {}).get('projects')]

    return [Project(**x) for x in data]

def get_user_commands() -> List[CommandBase.__subclasses__]:
    result = []
    for path in _load_config().get('COMMAND', []):
        cmd_cls = CommandCenter.register(path)
        if cmd_cls:
            result.append(cmd_cls)
    return result



from .base import *
from .project import Project

class Organization(ModelBase):
    '''
    {
      "isOrgAdmin": false,
      "name": "XXX公司",
      "id": "ffffffff"
    }
    '''

    def __init__(self, **kwargs: Any) -> None:
        projects: List[Project] = kwargs.pop("projects", [])
        super().__init__(**kwargs)
        self._projects: List[Project] = []
        self.projects = projects  # Use the setter to ensure binding

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Organization):
            return False
        return self.id == other.id

    @property
    def projects(self) -> List[Project]:
        """Return a copy of the projects list to prevent direct modification."""
        return self._projects.copy()

    @projects.setter
    def projects(self, value: List[Project]) -> None:
        """Bind projects to this organization when set."""
        self._projects = []
        for project in value:
            self.add_project(project)

    def add_project(self, project: Project) -> None:
        """Add a project to the organization and set its parent."""
        if project not in self._projects:
            self._projects.append(project)



from os.path import join
from typing import Tuple, List, Dict, Optional, Any, Union
import yaml
from yaml import Dumper, Loader

from ..sdk import OrganizationListAPI, ProjectListAPI
from ..model import ModelBase, Organization, Project, Member, Status
from .config import get_config_root, get_prefer_organization

def _setup_yaml_handlers():
    # Serialization: Convert ModelBase objects to YAML
    def model_base_representer(dumper: Dumper, obj: ModelBase) -> yaml.nodes.MappingNode:
        """Add a '__class__' field to track the object's type during serialization."""
        data = obj.dict.copy()
        data['__class__'] = obj.__class__.__name__  # Store class name
        return dumper.represent_mapping('!ModelBase', data)

    # Deserialization: Rebuild ModelBase objects from YAML
    def model_base_constructor(loader: Loader, node: yaml.nodes.MappingNode) -> ModelBase:
        """Use the '__class__' field to instantiate the correct subclass."""
        data = loader.construct_mapping(node)
        cls_name = data.pop('__class__')  # Extract class name
        cls = globals()[cls_name]         # Get the class from global scope
        return cls(**data)                # Rebuild the object

    # Register the custom handlers
    yaml.add_representer(ModelBase, model_base_representer)
    yaml.add_constructor('!ModelBase', model_base_constructor)

_setup_yaml_handlers()

def _get_config_file() -> str:
    return join(get_config_root(), 'cache.yaml')

def _save_to_cache(data: ModelBase) -> None:
    """Save a ModelBase object hierarchy to YAML."""
    with open(_get_config_file(), 'w') as f:
        yaml.dump(data, f, sort_keys=False)

def _load_from_cache() -> ModelBase:
    """Load a ModelBase object hierarchy from YAML."""
    with open(_get_config_file(), 'r') as f:
        return yaml.load(f, Loader=Loader)

def get_cached_organization(fetch: bool = False) -> Optional[Organization]:
    orgs = _load_from_cache()

    if fetch and not orgs:
        orgs = OrganizationListAPI.run()
        _save_to_cache(orgs)

        if not orgs:
            return None

    prefer_org_id = get_prefer_organization()

    if prefer_org_id:
        org = next((org for org in orgs if org.id == prefer_org_id))

        if org:
            return org

    return next((org for org in orgs if org.default == True), orgs[0])

def save_cached_organization(org: Organization):
    all_orgs = _load_from_cache()

    if all_orgs:
        # Filter out any existing org with same ID, then add the new one
        all_orgs = list(filter(lambda x: x != org, all_orgs))
        all_orgs.append(org)

        _save_to_cache(all_orgs)
    else:
        _save_to_cache([org])



from .base import *
from .member import Member
from .status import Status
from .workitemtype import WorkItemType

class Project(ModelBase):
    '''
    {
        'categoryIdentifier': 'Project',
        'creator': '1234567890',
        'customCode': 'ABC',
        'description': '',
        'gmtCreate': 1728538669000,
        'icon': 'https://img.alicdn.com/imgextra/i3/xxxxxxxx.png',
        'identifier': 'ffffffff',
        'name': 'NAME',
        'scope': 'private'
    }
    {
        'category': None,
        'categoryIdentifier': 'Project',
        'creator': '1234567890',
        'customCode': 'ABCD',
        'description': '',
        'gmtCreate': 1734185008000,
        'gmtModified': 1738839913000,
        'hasSuperiorSpace': True,
        'icon': 'https://img.alicdn.com/imgextra/i2/xxx.png',
        'iconBig': 'https://img.alicdn.com/imgextra/i1/xxx.png',
        'iconGroup': '{"small":"https://img.alicdn.com/imgextra/i2/xxx.png","big":"https://img.alicdn.com/imgextra/i1/xxx.png"}',
        'iconSmall': 'https://img.alicdn.com/imgextra/i2/xxx.png',
        'id': None,
        'identifier': 'ffffffff',
        'identifierPath': 'ffffffff',
        'logicalStatus': 'NORMAL',
        'modifier': '1234567890',
        'name': 'NAME',
        'organizationIdentifier': 'ffffffff',
        'parentIdentifier': None,
        'scope': 'public',
        'statusIdentifier': 'ffffffff',
        'statusStageIdentifier': '9',
        'subType': None,
        'typeIdentifier': 'AgileProject'
    }
    '''

    def __init__(self, **kwargs: Any) -> None:
        members: List[Member] = kwargs.pop("members", [])
        statuses: Dict[str, List[Status]] = kwargs.pop("statuses", {})
        workitemtypes: Dict[str, List[WorkItemType]] = kwargs.pop("workitemtypes", {})
        super().__init__(**kwargs)
        self._members: List[Member] = []
        self.members = members  # Use the setter to ensure binding
        self._statuses: Dict[str, List[Status]] = {}
        self.statuses = statuses  # Use the setter to ensure binding
        self._workitemtypes: Dict[str, List[WorkItemType]] = {}
        self.workitemtypes = workitemtypes  # Use the setter to ensure binding

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Project):
            return False
        return self.identifier == other.identifier

    @property
    def members(self) -> List[Member]:
        """Return a copy of the members list to prevent direct modification."""
        return self._members.copy()

    @members.setter
    def members(self, value: List[Member]) -> None:
        """Bind members to this project when set."""
        self._members = value

    @property
    def statuses(self) -> Dict[str, List[Status]]:
        """Return a copy of the statuses list to prevent direct modification."""
        return self._statuses.copy()

    @statuses.setter
    def statuses(self, value: Dict[str, List[Status]]) -> None:
        """Bind statuses to this project when set."""
        self._statuses = value

    def statuses_for(self, category: str) -> List[Status]:
        """Get the list of statuses for a specific work item category."""
        return self.statuses.get(category, [])

    def save_statuses_for(self, category: str, value: List[Status]) -> None:
        """Save the list of statuses for a specific work item category."""
        self._statuses[category] = value

    @property
    def workitemtypes(self) -> Dict[str, List[WorkItemType]]:
        """Return a copy of the workitem types list to prevent direct modification."""
        return self._workitemtypes.copy()

    @workitemtypes.setter
    def workitemtypes(self, value: Dict[str, List[WorkItemType]]) -> None:
        """Bind workitem types to this project when set."""
        self._workitemtypes = value

    def workitemtypes_for(self, category: str) -> List[WorkItemType]:
        """Get the list of workitem types for a specific work item category."""
        return self.workitemtypes.get(category, [])

    def save_workitemtypes_for(self, category: str, value: List[WorkItemType]) -> None:
        """Save the list of workitem types for a specific work item category."""
        self._workitemtypes[category] = value



from os.path import join
from typing import Tuple, List, Dict, Optional, Any, Union

from ..utils.config import get_config_root

_credential_file = join(get_config_root(), 'credential')

def get_login_ticket() -> Optional[str]:
    try:
        with open(_credential_file, 'r') as f:
            return f.read()
    except (FileNotFoundError, IOError):
        return None

def save_login_ticket(value: str):
    try:
        with open(_credential_file, 'w') as f:
            f.write(value)
    except IOError:
        print("Error saving the login ticket.")



from .organization_list import API as OrganizationListAPI
from .project_list import API as ProjectListAPI
from .project_detail import API as ProjectDetailAPI
from .member_list import API as MemberListAPI
from .workitem_list import API as WorkItemListAPI
from .workitem_detail import API as WorkItemDetailAPI
from .workitemfield import API as WorkItemFieldAPI
from .workitemtype import API as WorkItemTypeAPI
from .workflowstatus import API as WorkFlowStatusAPI

from .user_info import fetch_user_id

from .condition import Condition, ConditionGroup
from .category import Category



from .base import *

class Status(ModelBase):
    '''
    {
      "identifier": "100005",
      "creator": "AK-ADMIN",
      "workflowStageName": "确认阶段",
      "name": "待处理",
      "description": "",
      "source": "system",
      "gmtCreate": 1613805843000,
      "workflowStageIdentifier": "1",
      "resourceType": "Workitem"
    }
    '''

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Status):
            return False
        return self.identifier == other.identifier



from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from webdriver_manager.chrome import ChromeDriverManager
from typing import Optional
import time
import logging

def request_aliyun_login_ticket(headless: bool = False, timeout_minutes: int = 5) -> Optional[str]:
    """
    Open a webview using Selenium Chrome WebDriver and wait for user operations.
    
    Args:
        timeout_minutes (int): Maximum time to wait for user operations in minutes
        
    Returns:
        the aliyun login ticket
    """
    chrome_options = Options()
    if headless:
        chrome_options.add_argument("--headless")
    
    # Add additional options for better webview experience
    chrome_options.add_argument("--disable-gpu")
    chrome_options.add_argument("--no-sandbox")
    chrome_options.add_argument("--disable-dev-shm-usage")
    
    # Solution 1: Disable the Permissions Policy enforcement
    chrome_options.add_argument("--disable-blink-features=AutomationControlled")

    # # Solution 2: Explicitly allow sensor access
    # chrome_options.add_argument("--enable-features=WebXR")
    # chrome_options.add_argument("--enable-features=GenericSensorExtraClasses")
    # chrome_options.add_argument("--enable-features=WebXRHandInput")

    # # Alternative: Set Permissions Policy through experimental option
    # chrome_options.add_experimental_option("prefs", {
    #     "profile.default_content_setting_values.sensors": 1,
    #     "profile.default_content_setting_values.geolocation": 1
    # })

    login_ticket = None

    try:
        # Initialize the Chrome WebDriver
        driver = webdriver.Chrome(
            service=Service(ChromeDriverManager().install()),
            options=chrome_options
        )
        
        # Navigate to the initial URL
        initial_url = "https://devops.aliyun.com/workbench"
        driver.get(initial_url)
        logging.info(f"Opened initial URL: {initial_url}")

        # Calculate timeout in seconds
        timeout_seconds = timeout_minutes * 60
        start_time = time.time()
        last_url = initial_url

        while True:
            current_url = driver.current_url
            if current_url != last_url:
                logging.info(f"URL changed to: {current_url}")
                last_url = current_url

            if current_url == initial_url:
                logging.info('Login successfully')
                cookies = driver.get_cookies()

                cookie_ticket = next((cookie for cookie in cookies if cookie['name'] == 'login_aliyunid_ticket'), None)
                if cookie_ticket:
                    # logging.info(f"Found cookie: {cookie_ticket}")
                    login_ticket = cookie_ticket.get('value')

                break

            # Check if timeout has been reached
            if time.time() - start_time > timeout_seconds:
                logging.warning(f"Operation timeout reached after {timeout_minutes} minutes")
                break

            time.sleep(1)  # Check every second
    except Exception as e:
        logging.error(f"An error occurred: {str(e)}")
    finally:
        if 'driver' in locals():
            driver.quit()

    return login_ticket



from .base import *
from .workitemfield import WorkItemField

class WorkItemType(ModelBase):
    '''
    工作项类型

    {
      "systemDefault": true,
      "identifier": "xxxxxxxx",
      "creator": "AK-ADMIN",
      "enable": false,
      "addUser": "AK-ADMIN",
      "name": "任务",
      "description": "",
      "nameEn": "Task",
      "gmtAdd": 1728538669000,
      "categoryIdentifier": "Task",
      "gmtCreate": 1616584627000,
      "defaultType": false
    }
    '''

    def __init__(self, **kwargs: Any) -> None:
        fields: List[WorkItemField] = kwargs.pop("fields", [])
        super().__init__(**kwargs)
        self._fields: List[WorkItemField] = []
        self.fields = fields  # Use the setter to ensure binding

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, WorkItemType):
            return False
        return self.identifier == other.identifier

    @property
    def fields(self) -> List[WorkItemField]:
        """Return a copy of the fields to prevent direct modification."""
        return self._fields.copy()

    @fields.setter
    def fields(self, value: List[WorkItemField]) -> None:
        """Bind fields to this project when set."""
        self._fields = value



import click

from dataclasses import dataclass

from ..sdk import *
from .config import *
from .cache import *
from ..web import *

@dataclass
class GlobalState:
    _instance = None
    _organization: Organization = None 

    @classmethod
    def current(cls) -> 'GlobalState':
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    @property
    def user_id(self) -> str:
        '''Return the current user id.'''
        return get_aliyun_account_id() or fetch_user_id()

    @property
    def organization(self) -> Organization:
        '''Return the working on organization.'''
        if self._organization:
            return self._organization

        self._organization = get_cached_organization(True)

        if self._organization is None:
            raise click.exceptions.Exit(1)
        
        return self._organization

    @property
    def organization_id(self) -> str:
        '''Return the working on organization's id.'''
        return self.organization.id

    def save_organization(self):
        '''Save the working organization's changes to local cache.'''
        save_cached_organization(self.organization)

    def get_all_projects(self, reload: bool = False, reload_project_set: bool = False) -> List[Project]:
        '''Get all the projects under the working organization.'''
        all_projects = self.organization.projects

        if reload:
            # Fetch builtin projects
            all_projects = ProjectListAPI.run(self.organization_id)

            # Fetch the user's following projects
            for p in get_following_projects():
                if p not in all_projects:
                    all_projects.append(p)

            if reload_project_set:
                # Fetch from the user's project sets
                for p in self.get_all_projects_by_web():
                    if p not in all_projects:
                        all_projects.append(p)

            i, total = 1, len(all_projects)
            result_projects = []

            for p in all_projects:
                click.echo(f'Fetching {i}/{total}')
                i += 1

                click.echo(f'Fetching project details: {p.identifier}')
                project = ProjectDetailAPI.run(self.organization_id, p.identifier)

                click.echo(f'Fetching project members')
                members = MemberListAPI.run(self.organization_id, p.identifier)
                if members:
                    project.members = members

                click.echo(f'Fetching project workflow statuses')
                for category in list(Category):
                    statuses = WorkFlowStatusAPI.run(self.organization_id, p.identifier, category.value)
                    if statuses:
                        project.save_statuses_for(category.value, statuses)

                click.echo(f'Fetching project workitem\'s types and fields')
                for category in list(Category):
                    types = WorkItemTypeAPI.run(self.organization_id, p.identifier, category.value)

                    for t in types:
                        t.fields = WorkItemFieldAPI.run(self.organization_id, p.identifier, t.identifier)

                    project.save_workitemtypes_for(category.value, types)

                result_projects.append(project)
                click.secho(f'Reloaded project {project.name}\n', fg='green', bold=True)

            all_projects = result_projects
            self.organization.projects = all_projects
            self.save_organization()

        return all_projects

    def get_all_projects_by_web(self) -> List[Project]:
        ticket = get_login_ticket()

        if ticket:
            data = fetch_user_info(ticket)

            if data:
                if data.get('tenantId') != self.organization_id:
                    click.echo('Not a matching organization, you should switch your organization to prefer one and try again.')
                    click.echo(f'{data}')
                    return
            else:
                click.echo(f'Login ticket may get expired: {ticket}')
                ticket = None

        if ticket is None:
            click.echo('Opening chrome browser to get a access ticket of aliyun...')
            ticket = request_aliyun_login_ticket()

            if ticket:
                save_login_ticket(ticket)

        if not ticket:
            return
        
        all_projects = []
        all_project_sets = fetch_project_sets(ticket) or []

        if len(all_project_sets) > 0:
            for projectset in all_project_sets:
                id = projectset.get('identifier')
                name = projectset.get('name')
                click.echo(f'Fetching project set {name} ({id})')
                all_projects.extend(fetch_projects(id, ticket))
        else:
            click.echo('No project set found')
        
        return all_projects

    def get_matching_members(self, user) -> List[Member]:
        results = []

        for project in self.organization.projects:
            for member in project.members:
                if member.identifier == user or member.realName == user:
                    results.append(member)
        
        return results

    def get_matching_member_name(self, user) -> str:
        members = self.get_matching_members(user)
        return members[0].realName if len(members) > 0 else user

    def get_matching_status(self, name) -> List[Status]:
        results = []

        for project in self.organization.projects:
            for _, statuses in project.statuses.items():
                for status in statuses:
                    if status.name == name:
                        results.append(status)
        
        return results

    def get_matching_status_ids(self, keywords) -> List[str]:
        if not keywords:
            return []

        results = []
        names = keywords.split(',')

        for project in self.organization.projects:
            for _, statuses in project.statuses.items():
                for status in statuses:
                    if status.name in names:
                        results.append(status)
        
        return list(map(lambda x: x.identifier, results))
    
    def get_matching_field_name(self, project_id: str, category_id: str, field_id: str) -> str:
        for project in self.organization.projects:
            if project.identifier != project_id:
                continue

            for category, workitemtypes in project.workitemtypes.items():
                if category != category_id:
                    continue

                for workitemtype in workitemtypes:
                    for field in workitemtype.fields:
                        if field.identifier == field_id:
                            return field.name
        
        return None

    def get_matching_field_id(self, project_id: str, category_id: str, field_name: str) -> str:
        for project in self.organization.projects:
            if project.identifier != project_id:
                continue

            for category, workitemtypes in project.workitemtypes.items():
                if category != category_id:
                    continue

                for workitemtype in workitemtypes:
                    for field in workitemtype.fields:
                        if field.name == field_name:
                            return field.identifier
        
        return None


# BEGIN - Functions

def list_workitem_by(category: Category, assigned_to: Optional[str], statuses: Optional[str] = None, not_in_statuses: Optional[str] = None):
    '''List workitem by conditions.'''

    group = ConditionGroup()

    if assigned_to:
        user_id = next(map(lambda x: x.identifier, GlobalState.current().get_matching_members(assigned_to)))
    else:
        user_id = GlobalState.current().user_id

    if user_id:
        group.add(Condition.assigned_to(user_id))

    statuses = GlobalState.current().get_matching_status_ids(statuses)
    not_in_statuses = GlobalState.current().get_matching_status_ids(not_in_statuses)

    if statuses and len(statuses) > 0:
        group.add(Condition.status_contains(statuses))

    if not_in_statuses and len(not_in_statuses) > 0:
        group.add(Condition.status_not_contains(not_in_statuses))

    for project in GlobalState.current().get_all_projects():
        workitems = WorkItemListAPI.run(GlobalState.current().organization_id, project.identifier, category.value, group.dict)
        yield project, workitems

# END - Functions


from datetime import datetime

def format_timestamp(ts: int) -> str:
    # Convert from microseconds to seconds by dividing by 1,000,000
    dt = datetime.fromtimestamp(ts / 1_000_000)
    return dt.strftime('%Y-%m-%d')



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="yunxiao-cli",
    version="0.1.0",
    author="Wei Han",
    author_email="xingheng.hax@qq.com",
    description="A CLI tool for interacting with Alibaba Cloud Yunxiao",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/python-cli/yunxiao-cli",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.7",
    install_requires=[
        "alibabacloud-tea",
        "alibabacloud-tea-openapi",
        "aliyun-python-sdk-core",
        "aliyun-python-sdk-sts",
        "aliyunsdkcore",
        "click",
        "PyYAML",
        "requests",
        "rich",
        "selenium",
        "webdriver-manager",
    ],
    entry_points={
        "console_scripts": [
            "yunxiao-cli=yunxiao.main:cli",
        ],
    },
)



from aliyunsdkcore.client import AcsClient
from aliyunsdksts.request.v20150401 import GetCallerIdentityRequest
from json import loads

from ..utils.config import *

def fetch_user_id():
    '''
    Return current user's aliyun account ID.
    '''
    id, secret = get_credential()
    client = AcsClient(id, secret, 'cn-hangzhou')
    request = GetCallerIdentityRequest.GetCallerIdentityRequest()
    response = client.do_action_with_exception(request)
    
    if not response:
        return None

    data = loads(response.decode('utf-8'))
    return data.get('UserId')



from typing import List, Dict, Any
from abc import ABC, abstractmethod
from json import dumps, loads

from alibabacloud_tea_openapi.client import Client as OpenApiClient
from alibabacloud_tea_openapi import models as open_api_models
from alibabacloud_tea_util import models as util_models
from alibabacloud_openapi_util.client import Client as OpenApiUtilClient

from ..utils.config import get_credential, get_endpoint

#
# Docs:
# https://api.aliyun.com/api/devops/
#

class APIBase(ABC):
    """Abstract base class for API endpoints"""
    
    @staticmethod
    def create_client() -> OpenApiClient:
        """
        使用AK&SK初始化账号Client
        @return: Client
        @throws Exception
        """
        id, secret = get_credential()
        config = open_api_models.Config(access_key_id=id, access_key_secret=secret, endpoint=get_endpoint())
        return OpenApiClient(config)

    @staticmethod
    @abstractmethod
    def create_api_info(**kwargs) -> open_api_models.Params:
        """
        API 相关
        @param path: string Path parameters
        @return: OpenApi.Params
        """
        pass

    @staticmethod
    @abstractmethod
    def run(**kwargs) -> Any | None:
        """
        Send API request.
        """
        pass



from .base import *
from .category import Category
from ..model import WorkItem

class API(APIBase):

    @staticmethod
    def create_api_info(
        organization_id: str,
    ) -> open_api_models.Params:
        """
        API 相关
        @param path: string Path parameters
        @return: OpenApi.Params
        """
        params = open_api_models.Params(
            # 接口名称,
            action='ListWorkitems',
            # 接口版本,
            version='2021-06-25',
            # 接口协议,
            protocol='HTTPS',
            # 接口 HTTP 方法,
            method='GET',
            auth_type='AK',
            style='ROA',
            # 接口 PATH,
            pathname=f'/organization/{organization_id}/listWorkitems',
            # 接口请求体内容格式,
            req_body_type='json',
            # 接口响应体内容格式,
            body_type='json'
        )
        return params

    @staticmethod
    def run(
        organization: str,
        project: str,
        category: str,
        condition: Dict | None = None,
    ) -> List[WorkItem]:
        client = API.create_client()
        params = API.create_api_info(organization)
        # query params
        queries = {}
        queries['spaceType'] = 'Project'
        queries['spaceIdentifier'] = project
        queries['category'] = category
        queries['conditions'] = dumps(condition) if condition else None
        # runtime options
        runtime = util_models.RuntimeOptions()
        request = open_api_models.OpenApiRequest(
            query=OpenApiUtilClient.query(queries)
        )
        # 复制代码运行请自行打印 API 的返回值
        # 返回值实际为 Map 类型，可从 Map 中获得三类数据：响应体 body、响应头 headers、HTTP 返回的状态码 statusCode。
        dict = client.call_api(params, request, runtime)
        data = dict.get('body', {}).get('workitems') or []
        return [WorkItem(**d) for d in data]



import requests
from typing import Optional, Dict, List
from ..model.project import Project

def fetch_user_info(ticket: str) -> Optional[Dict]:
    url = 'https://devops.aliyun.com/uiless/api/sdk/users/me'
    cookies = {
        'login_aliyunid_ticket': ticket
    }

    response = requests.get(url, cookies=cookies)
    data = response.json()

    if data.get('code') != 200:
        print(f'Failed to get user info: {data}')
        return None

    return data.get('result')

def fetch_project_sets(ticket: str) -> Optional[Dict]:
    url = 'https://devops.aliyun.com/projex/api/workspace/program/search/list'
    params = {
        'category': 'Program',
        'toPage': 1,
        'pageSize': 20,
        '_input_charset': 'utf-8'
    }
    headers = {
        'content-type': 'application/json'
    }
    cookies = {
        'login_aliyunid_ticket': ticket
    }

    response = requests.get(url, headers=headers, cookies=cookies, params=params)
    data = response.json()

    if data.get('code') != 200:
        print(f'Failed to get project sets: {data}')
        return None

    return data.get('result')

def fetch_projects(project_set_id: str, ticket: str) -> List[Project]:
    url = f'https://devops.aliyun.com/projex/api/workspace/program/{project_set_id}/binding/project/list'
    headers = {
        'content-type': 'application/json',
    }
    cookies = {
        'login_aliyunid_ticket': ticket
    }

    response = requests.get(url, headers=headers, cookies=cookies)
    data = response.json()

    if data.get('code') != 200:
        print(f'Failed to get project sets: {data}')
        return None

    return [Project(**d) for d in data.get('result', [])]
